@@ -22,7 +22,16 @@ from functools import partial
 import itertools
 import json
 from pathlib import Path
-from numba import jit
+try:
+    from numba import jit, njit
+except ImportError:
+    # Some systems may not have numba capabilities
+    def jit(*args, **kwargs):
+        """Dummy decorator, does nothing"""
+        def _decorator(func):
+            return func
+        return _decorator
+
 from numbers import Number
 import numpy as np
 import struct
@@ -46,20 +55,49 @@ except ImportError:
     from yaml import Dumper as YAMLDumper
 
 
-def read_pfb(file: str, mode: str='full', z_first: bool=True):
+def read_pfb(file: str, keys: dict=None, mode: str='full', z_first: bool=True,
+             read_sg_info: bool=True):
     """
     Read a single pfb file, and return the data therein
 
     :param file:
         The file to read.
+    :param keys:
+        A set of keys for indexing subarrays of the full pfb. Optional.
+        This is mainly a trick for interfacing with xarray, but the format
+        of the keys is:
+
+            ::
+            {'x': {'start': start_x, 'stop': end_x},
+             'y': {'start': start_y, 'stop': end_y},
+             'z': {'start': start_z, 'stop': end_z}}
+
     :param mode:
         The mode for the reader. See ``ParflowBinaryReader::read_all_subgrids``
         for more information about what modes are available.
+    :param read_sg_info:
+        Precalculating subgrid information does not always work correctly for
+        some files, especially velocity files. If ``True``, read subgrid info 
+        directly from the pfb file (slower, but more robust)
     :return:
         An nd array containing the data from the pfb file.
     """
-    with ParflowBinaryReader(file) as pfb:
-        data = pfb.read_all_subgrids(mode=mode, z_first=z_first)
+    with ParflowBinaryReader(file, read_sg_info=read_sg_info) as pfb:
+        if not keys:
+            data = pfb.read_all_subgrids(mode=mode, z_first=z_first)
+        else:
+            base_header = pfb.header
+            start_x = keys.get('x', {}).get('start', None) or 0
+            start_y = keys.get('y', {}).get('start', None) or 0
+            start_z = keys.get('z', {}).get('start', None) or 0
+            stop_x =  keys.get('x', {}).get('stop', None) or base_header['nx']
+            stop_y =  keys.get('y', {}).get('stop', None) or base_header['ny']
+            stop_z =  keys.get('z', {}).get('stop', None) or base_header['nz']
+            nx = np.max([stop_x - start_x, 1])
+            ny = np.max([stop_y - start_y, 1])
+            nz = np.max([stop_z - start_z, 1])
+            data = pfb.read_subarray(
+                        start_x, start_y, start_z, nx, ny, nz, z_first=z_first)
     return data
 
 
@@ -205,7 +243,8 @@ def read_pfb_sequence(
     file_seq: Iterable[str],
     keys=None,
     z_first: bool=True,
-    z_is: str='z'
+    z_is: str='z',
+    read_sg_info: bool=False
 ):
     """
     An efficient wrapper to read a sequence of pfb files. This
@@ -231,13 +270,17 @@ def read_pfb_sequence(
     :param z_is:
         A descriptor of what the z axis represents. Can be one of
         'z', 'time', 'variable'. Default is 'z'.
-
+    :param read_sg_info:
+        Precalculating subgrid information does not always work correctly for
+        some files, especially velocity files. If ``True``, read subgrid info 
+        directly from the pfb file (slower, but more robust)
+    
     :return:
         An nd array containing the data from the files.
     """
     # Filter out unique files only
     file_seq = sorted(list(set(file_seq)))
-    with ParflowBinaryReader(file_seq[0]) as pfb_init:
+    with ParflowBinaryReader(file_seq[0], read_sg_info=read_sg_info) as pfb_init:
         base_header = pfb_init.header
         base_sg_offsets = pfb_init.subgrid_offsets
         base_sg_locations = pfb_init.subgrid_locations
@@ -248,16 +291,17 @@ def read_pfb_sequence(
     if not keys:
         nx, ny, nz = base_header['nx'], base_header['ny'], base_header['nz']
     else:
-        start_x = keys['x']['start'] or 0
-        start_y = keys['y']['start'] or 0
-        start_z = keys[z_is]['start'] or 0
-        stop_x = keys['x']['stop'] or base_header['nx']
-        stop_y = keys['y']['stop'] or base_header['ny']
-        stop_z = keys[z_is]['stop'] or base_header['nz']
+        start_x = keys.get('x', {}).get('start', None) or 0
+        start_y = keys.get('y', {}).get('start', None) or 0
+        start_z = keys.get(z_is, {}).get('start', None) or 0
+        stop_x =  keys.get('x', {}).get('stop', None) or base_header['nx']
+        stop_y =  keys.get('y', {}).get('stop', None) or base_header['ny']
+        stop_z =  keys.get(z_is, {}).get('stop', None) or base_header['nz']
         nx = np.max([stop_x - start_x, 1])
         ny = np.max([stop_y - start_y, 1])
         nz = np.max([stop_z - start_z, 1])
 
+    n_seq = len(file_seq)
     if z_first:
         seq_size = (len(file_seq), nz, ny, nx)
     else:
@@ -325,6 +369,13 @@ class ParflowBinaryReader:
     :param header:
         A dictionary representing the header of the pfb file. This is an optional
         input, if it is not given we will read it from the pfb file directly.
+    :param read_sg_info:
+        Whether or not to read subgrid information directly from the pfb file.
+        Precalculating subgrid information does not always work correctly for
+        some files, especially velocity files. This reads the subgrid offset 
+        bytes, subgrid indices and subgrid shapes, then computes the subgrid 
+        coordinates subgrid locations, and chunk sizes as normal
+        
     """
 
     def __init__(
@@ -334,7 +385,8 @@ class ParflowBinaryReader:
         p: int=None,
         q: int=None,
         r: int=None,
-        header: Mapping[str, Number]=None
+        header: Mapping[str, Number]=None,
+        read_sg_info: bool=False
     ):
         self.filename = file
         self.f = open(self.filename, 'rb')
@@ -342,15 +394,15 @@ class ParflowBinaryReader:
             self.header = self.read_header()
         else:
             self.header = header
-        self.header['p'] = self.header.get('p', p)
-        self.header['q'] = self.header.get('q', q)
-        self.header['r'] = self.header.get('r', r)
 
         if np.all([p, q, r]):
             self.header['p'] = p
             self.header['q'] = q
             self.header['r'] = r
-        else:
+
+        if not ('p' in self.header
+            and 'q' in self.header
+            and 'r' in self.header):
             # If p, q, and r aren't given we can precompute them
             # NOTE: This is a bit of a fallback and may not always work
             eps = 1 - 1e-6
@@ -362,6 +414,9 @@ class ParflowBinaryReader:
         if precompute_subgrid_info:
             self.compute_subgrid_info()
 
+        if read_sg_info:
+            self.read_subgrid_info()
+            
     def close(self):
         self.f.close()
 
@@ -391,6 +446,41 @@ class ParflowBinaryReader:
         self.chunks = self._compute_chunks()
         self.coords = self._compute_coords()
 
+    def read_subgrid_info(self):
+        """ Read the header for each subgrid directly from the pfb file, rather 
+        than calculating it via precalculate_subgrid_info """
+        
+        sg_shapes = []
+        sg_offs = []
+        sg_locs = []
+        sg_starts = []
+
+        off = 64
+        for sg_num in range(self.header['n_subgrids']):
+            # Read and move past the current subgrid header
+            sg_head = self.read_subgrid_header(off)
+            off += 36 
+            
+            sg_starts.append([sg_head['ix'], sg_head['iy'], sg_head['iz']])
+            sg_shapes.append([sg_head['nx'], sg_head['ny'], sg_head['nz']])
+            sg_offs.append(off)
+
+            # Calculate subgrid locs instead of reading from file
+            sg_p, sg_q, sg_r = get_subgrid_loc(sg_num, self.header['p'], 
+                                                       self.header['q'], 
+                                                       self.header['r'])
+            sg_locs.append((sg_p, sg_q, sg_r))
+            
+            # Finally, move past the current subgrid before next iteration
+            off += sg_head['sg_size']*8
+        
+        self.subgrid_offsets = np.array(sg_offs)
+        self.subgrid_locations = np.array(sg_locs)
+        self.subgrid_start_indices = np.array(sg_starts)
+        self.subgrid_shapes = np.array(sg_shapes)
+        self.chunks = self._compute_chunks()
+        self.coords = self._compute_coords()
+
     def _compute_chunks(self) -> Mapping[str, tuple]:
         """
         This computes the chunk sizes of the subgrids. Note that it does
@@ -455,15 +545,23 @@ class ParflowBinaryReader:
         """Reads a subgrid header at the position ``skip_bytes``"""
         self.f.seek(skip_bytes)
         sg_header = {}
-        sg_header['ix'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['iy'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['iz'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['nx'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['ny'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['nz'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['rx'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['ry'] = struct.unpack('>i', self.f.read(4))[0]
-        sg_header['rz'] = struct.unpack('>i', self.f.read(4))[0]
+        header_len = 9
+
+        # Apologies, this is kind of ugly, but faster than using struct
+        (sg_header['ix'],
+         sg_header['iy'],
+         sg_header['iz'],
+         sg_header['nx'],
+         sg_header['ny'],
+         sg_header['nz'],
+         sg_header['rx'],
+         sg_header['ry'],
+         sg_header['rz']) = np.memmap(self.f,
+                                      dtype=np.int32,
+                                      offset=skip_bytes,
+                                      mode='r',
+                                      shape=(header_len,),
+                                      order='F').byteswap()
         sg_header['sg_size'] = np.prod([sg_header[n] for n in ['nx', 'ny', 'nz']])
         return sg_header
 
@@ -535,8 +633,19 @@ class ParflowBinaryReader:
                 if end in c: break
             return np.arange(s, e+1)
 
+        if not start_x:
+            start_x = 0
+        if not start_y:
+            start_y = 0
+        if not start_z:
+            start_z = 0
+        if not nx:
+            nx = self.header['nx']
+        if not ny:
+            ny = self.header['ny']
         if not nz:
             nz = self.header['nz']
+
         end_x = start_x + nx
         end_y = start_y + ny
         end_z = start_z + nz
@@ -1193,14 +1302,16 @@ def _read_vegm(file_name):
     x_dim = int(last_line_split[0])
     y_dim = int(last_line_split[1])
     z_dim = len(last_line_split) - 2
-    vegm_array = np.zeros((x_dim, y_dim, z_dim))
+
+    # To be consistent with ParFlow-python xy indexing, x should represent columns and y rows
+    vegm_array = np.zeros((y_dim, x_dim, z_dim))
     # Assume first two lines are comments
     for line in lines[2:]:
         elements = line.split()
         x = int(elements[0])
         y = int(elements[1])
         for i in range(z_dim):
-            vegm_array[x - 1, y - 1, i] = elements[i + 2]
+            vegm_array[y - 1, x - 1, i] = elements[i + 2]
 
     return vegm_array
 
@@ -1272,18 +1383,7 @@ class DataAccessor:
     # ---------------------------------------------------------------------------
 
     def _pfb_to_array(self, file_path):
-        from parflowio.pyParflowio import PFData
-
-        array = None
-        if file_path:
-            full_path = get_absolute_path(file_path)
-            # FIXME do something with selector inside parflow-io
-            pfb_data = PFData(full_path)
-            pfb_data.loadHeader()
-            pfb_data.loadData()
-            array = pfb_data.moveDataArray()
-
-        return array
+        return read_pfb(get_absolute_path(file_path))
 
     # ---------------------------------------------------------------------------
     # time
